/* 
 * $Id: route_atob.c,v 1.9 1998/11/23 22:50:21 masaki Exp $
 */

#include <mrt.h>
#include <bgp.h>
#include <ctype.h>

static void read_ascii (FILE *fd, io_t * io);
static void skip (FILE * fd);

#ifndef HAVE_STRPTIME
char *strptime (char *buf, const char *format, struct tm *tm);
#endif /* HAVE_STRPTIME */


void
main (int argc, char *argv[])
{
    int c;
    extern char *optarg;	/* getopt stuff */
    extern int optind;		/* getopt stuff */
    char *usage =
    "Usage: route_atob [-i ascii_data_file] [-(o|w) binary_out_file]\n";
    char *filename = "stdin";
    int errors = 0;
    io_t *IO;
    trace_t *default_trace;
    FILE *fd;

    default_trace = New_Trace ();
    IO = New_IO (default_trace);
    io_set (IO, IO_OUTFILE, "stdout", NULL);

    while ((c = getopt (argc, argv, "i:f:o:w:v")) != -1)
	switch (c) {
	case 'o':
	    if (io_set (IO, IO_OUTFILE, (char *) (optarg), NULL) < 0) {
		printf ("Error setting outfile\n");
		errors++;
	    }
	    break;
	case 'w':
	    if (io_set (IO, IO_OUTMSGQ, (char *) (optarg), NULL) < 0) {
		printf ("Error setting output %s\n", optarg);
		errors++;
	    }
	    break;
	case 'i':
	case 'f':
	    filename = (char *) optarg;
	    break;
	case 'v':
	    set_trace (default_trace, TRACE_FLAGS, TR_ALL,
		       TRACE_LOGFILE, "stdout", NULL);
	    break;
	default:
	    errors++;
	}

    if (errors) {
	fprintf (stderr, usage);
	printf ("\nMRT version (%s) compiled on %s\n\n",
		MRT_VERSION, __DATE__);
	exit (0);
    }

    init_mrt (default_trace);

    if (!strcasecmp (filename, "stdin"))
	fd = stdin;
    else
	fd = fopen (filename, "r");

    if (fd == NULL) {
	perror (filename);
	exit (1);
    }
    read_ascii (fd, IO);
    exit (0);
}


static char *
checkstradvance (char *cp, char *s)
{
    int len = strlen (s);
    if (strncasecmp (cp, s, len) == 0) {
	if (cp[len] == '\0') {
	    return (cp + len);
	}
	if (cp[len] == ':' || isspace (cp[len])) {
	    do {
		len++;
	    } while (isspace (cp[len]));
	    return (cp + len);
	}
	return (NULL);
    }
    return (NULL);
}


static int
nulline (char *cp)
{
    while (isspace (*cp))
	cp++;
    if (*cp == '\0')
	return (1);
    return (0);
}


static void
flushout (io_t *io, time_t tstamp, int type, int bgp_type, bgp_attr_t *attr, 
	  LINKED_LIST *ll_ann, LINKED_LIST *ll_with, gateway_t *gateway_to)
{
    u_char *cp;
    int size;
    char *stime, **cpp;

    stime = my_strftime (tstamp, "%D %T");
    trace (TR_TRACE, MRT->trace, "TIME: %s\n", stime);
    if ((cpp = S_MRT_MSG_SUBTYPES[type]) != NULL)
        trace (TR_TRACE, MRT->trace, "TYPE: %s/%s\n", 
	       S_MRT_MSG_TYPES[type], cpp[bgp_type]);
    else
        trace (TR_TRACE, MRT->trace, "TYPE: %s\n", S_MRT_MSG_TYPES[type]);
    Delete (stime);

    if (attr->gateway)
        trace (TR_TRACE, MRT->trace, "FROM: %s AS%d\n",
	       prefix_toa (attr->gateway->prefix), attr->gateway->AS);
    if (gateway_to) {
        trace (TR_TRACE, MRT->trace, "TO: %s AS%d\n", 
	       prefix_toa (gateway_to->prefix), gateway_to->AS);
    }
    cp = bgp_create_update_msg (type, &size, attr, ll_ann, ll_with, gateway_to);
    io_write (io, tstamp, type, bgp_type, size, (char *) cp);
    Delete (cp);
    trace (TR_TRACE, MRT->trace, "write %d bytes\n", size);
    trace (TR_TRACE, MRT->trace, "\n");
}


static void 
read_ascii (FILE *fd, io_t *IO)
{
    char line[MAXLINE], *ret;
    bgp_attr_t *attr;
    int announce_flag;
    LINKED_LIST *ll_ann, *ll_with;
    int ok;
    int line_num = 0;
    int have_data;
    time_t tstamp;
    enum MRT_MSG_TYPES type;
    enum MRT_MSG_BGP_TYPES bgp_type;
    gateway_t *gateway_to;
    int first;

    announce_flag = 0;
    have_data = 0;
    tstamp = 0;
    type = 0;
    bgp_type = 0;
    gateway_to = NULL;
    first = 1;

    attr = bgp_new_attr (PROTO_BGP);
    ll_ann = LL_Create (LL_DestroyFunction, Deref_Prefix, 0);
    ll_with = LL_Create (LL_DestroyFunction, Deref_Prefix, 0);

    while (fgets (line, MAXLINE - 1, fd)) {

	if (first && nulline (line))
	    continue;
	first = 0;

	line[strlen (line) - 1] = '\0';
	trace (TR_TRACE, MRT->trace, "++%s\n", line);
        line_num++;

	if ((ret = checkstradvance (line, "TIME"))) {
	    struct tm tm;
	    if (strptime (ret, "%D %T", &tm)) {
		time_t now;
		time (&now);
		tm.tm_isdst = localtime (&now)->tm_isdst;
		tstamp = mktime (&tm);
	    }
	}

	else if ((ret = checkstradvance (line, "TO"))) {
	    int as;
    	    char data[MAXLINE];
	    prefix_t *prefix;
	    if (sscanf (ret, "%s AS%d", data, &as) >= 2 ||
	        sscanf (ret, "AS%d %s", &as, data) >= 2 /* obsolete */) {
		prefix = ascii2prefix (0, data);
		gateway_to = add_gateway (prefix, as, NULL);
		Deref_Prefix (prefix);
	    }
	}

	else if ((ret = checkstradvance (line, "FROM"))) {
	    int as;
    	    char data[MAXLINE];
	    if (sscanf (ret, "%s AS%d", data, &as) >= 2 ||
	        sscanf (ret, "AS%d %s", &as, data) >= 2 /* obsolete */) {
		prefix_t *prefix = ascii2prefix (0, data);
		attr->gateway = add_gateway (prefix, as, NULL);
		Deref_Prefix (prefix);
	    }
	}

	/* type BGP/UPDATE */
	else if ((ret = checkstradvance (line, "TYPE"))) {
	    char *subtypestr;
	    char **cpp;
	    int subtype = 0;
	    int i;

	    if ((subtypestr = strchr (ret, '/')) != NULL)
		*subtypestr++ = '\0';

	    cpp = S_MRT_MSG_TYPES;
	    for (i = 0; cpp[i]; i++) {
	        if (strcasecmp (cpp[i], ret) == 0)
			break;
	    }
	    type = (enum MRT_MSG_TYPES) i;

	    if (subtypestr) {
		if ((cpp = S_MRT_MSG_SUBTYPES[type]) != NULL) {
		    for (i = 0; cpp[i]; i++) {
		        if (strcasecmp (cpp[i], subtypestr) == 0)
			    break;
		    }
		    subtype = i;
		}
	    }

	    if (type == MSG_PROTOCOL_BGP
#ifdef HAVE_IPV6
		    || type == MSG_PROTOCOL_BGP4PLUS
		    || type == MSG_PROTOCOL_BGP4PLUS_01
#endif /* HAVE_IPV6 */
		    ) {
		bgp_type = (enum MRT_MSG_BGP_TYPES) subtype;

		if (bgp_type == MSG_BGP_UPDATE) {
		    have_data++;
		}
		else {
		    trace (TR_ERROR, MRT->trace, 
			   "Unknown message subtype %s at line %d\n",
			   subtypestr, line_num);
		    /* skip (fd); */
		}
	    }
	    else {
		trace (TR_ERROR, MRT->trace, 
		       "Unknown message type %s at line %d\n", ret, line_num);
		/* skip (fd); */
	    }
	}

	else if ((ret = checkstradvance (line, "ANNOUNCE"))) {
	    announce_flag = 1;
	}
	else if ((ret = checkstradvance (line, "WITHDRAW"))) {
	    announce_flag = 0;
	}
	else if ((ok = bgp_scan_attr (line, attr, MRT->trace)) > 0) {
	    /* OK */
	}
	else if (ok < 0) {
	    trace (TR_ERROR, MRT->trace, 
	           "Invalid BGP attribute at line %d\n", line_num);
	}

	else if (nulline (line)) {
	    if (have_data) {
		flushout (IO, tstamp, type, bgp_type, attr, ll_ann, ll_with, 
			  gateway_to);
	    }

	    bgp_deref_attr (attr);
	    attr = bgp_new_attr (PROTO_BGP);
	    LL_Clear (ll_ann);
	    LL_Clear (ll_with);

    	    announce_flag = 0;
    	    have_data = 0;
    	    tstamp = 0;
    	    type = 0;
    	    bgp_type = 0;
    	    gateway_to = NULL;
    	    first = 1;
	}
	/* a prefix -- line begins with a space */
	else if (isspace (line[0])) {
	    prefix_t *prefix;
	    char *cp = line +1;

	    while (isspace (*cp))
		cp++;
	    if ((prefix = ascii2prefix (0, cp)) == NULL) {
		trace (TR_ERROR, MRT->trace, 
		       "Unrecognized prefix at line %d\n", line_num);
	    }
	    else {
		if (announce_flag == 1) {
		    LL_Add (ll_ann, prefix);
		}
		else {
		    LL_Add (ll_with, prefix);
		}
	    }
	}
	else {
	    trace (TR_ERROR, MRT->trace, "Unrecognized line at %d\n", line_num);
	}
    }

    if (have_data) {
	flushout (IO, tstamp, type, bgp_type, attr, ll_ann, ll_with, 
		  gateway_to);
    }
    bgp_deref_attr (attr);
    LL_Destroy (ll_ann);
    LL_Destroy (ll_with);
}


static void 
skip (FILE * fd)
{
    char line[MAXLINE];

    while (fgets (line, MAXLINE - 1, fd) != NULL) {
	if (line[0] == '\n')
	    return;
    }
    return;
}


#ifndef HAVE_STRPTIME

/*
 * Please be careful because this function doesn't consult the format.
 * This only parses "MM/DD/YY HH:MM:SS" like 05/06/96 23:55:10.
 */

char *
strptime (char *buf, const char *format, struct tm *tm)
{

    static char *tm_zone = NULL;
    static long tm_gmtoff = 0;
    static int tm_isdst = 0;

    if (strlen (buf) < 17)
	return (NULL);

    if (tm_zone == NULL && tm_gmtoff == 0) {
	struct tm *tmp;
	time_t now;
	time (&now);
	tmp = localtime (&now);
	tm_isdst = tmp->tm_isdst;
	tm_gmtoff = tmp->tm_gmtoff;
	tm_zone = strdup (tmp->tm_zone);
    }
    buf[2] = buf[5] = buf[8] = buf[11] = buf[14] = '\0';
    tm->tm_mon = atoi (buf + 0) - 1;
    tm->tm_mday = atoi (buf + 3);
    tm->tm_year = atoi (buf + 6);
    tm->tm_hour = atoi (buf + 9);
    tm->tm_min = atoi (buf + 12);
    tm->tm_sec = atoi (buf + 15);
    tm->tm_gmtoff = tm_gmtoff;
    tm->tm_zone = tm_zone;

    return (buf + 17);
}

#endif /* HAVE_STRPTIME */
